/*
 *	clipmath.c
 *
 *	CLiP - Common Library for P/ECE
 *	Copyright (C) 2001 Naoyuki Sawa
 *
 *	* Sat Dec  1 13:20:55 JST 2001 Naoyuki Sawa
 *	- 쐬JnB
 */
#include "clip.h"

/*****************************************************************************
 *	xNg
 *****************************************************************************/

/* L@̒` */
#define Ax	a->x
#define Ay	a->y
#define Az	a->z
#define Bx	b->x
#define By	b->y
#define Bz	b->z

const VECTOR VECTOR_0 = { 0, 0, 0 };
const VECTOR VECTOR_X = { 1, 0, 0 };
const VECTOR VECTOR_Y = { 0, 1, 0 };
const VECTOR VECTOR_Z = { 0, 0, 1 };

VECTOR*
vector_init(VECTOR* o, float x, float y, float z)
{
	o->x = x;
	o->y = y;
	o->z = z;
	return o;
}

VECTOR*
vector_neg(VECTOR* o, const VECTOR* a)
{
	return vector_init(o,
		-Ax,
		-Ay,
		-Az);
}

VECTOR*
vector_add(VECTOR* o, const VECTOR* a, const VECTOR* b)
{
	return vector_init(o,
		Ax + Bx,
		Ay + By,
		Az + Bz);
}

VECTOR*
vector_sub(VECTOR* o, const VECTOR* a, const VECTOR* b)
{
	return vector_init(o,
		Ax - Bx,
		Ay - By,
		Az - Bz);
}

VECTOR*
vector_mul(VECTOR* o, const VECTOR* a, const VECTOR* b)
{
	return vector_init(o,
		Ax * Bx,
		Ay * By,
		Az * Bz);
}

VECTOR*
vector_div(VECTOR* o, const VECTOR* a, const VECTOR* b)
{
	return vector_init(o,
		Ax / Bx,
		Ay / By,
		Az / Bz);
}

VECTOR*
vector_mul1(VECTOR* o, const VECTOR* a, float s)
{
	return vector_init(o,
		Ax * s,
		Ay * s,
		Az * s);
}

VECTOR*
vector_div1(VECTOR* o, const VECTOR* a, float s)
{
	return vector_init(o,
		Ax / s,
		Ay / s,
		Az / s);
}

float
vector_dot(const VECTOR* a, const VECTOR* b)
{
	return Ax * Bx +
	       Ay * By +
	       Az * Bz;
}
VECTOR*
vector_cross(VECTOR* o, const VECTOR* a, const VECTOR* b)
{
	return vector_init(o,
		Ay * Bz - Az * By,
		Az * Bx - Ax * Bz,
		Ax * By - Ay * Bx);
}

VECTOR*
vector_normal(VECTOR* o, const VECTOR* a)
{
	return vector_div1(o, a, vector_mag(a));
}

float
vector_mag(const VECTOR* a)
{
	return sqrt(vector_mag2(a));
}

float
vector_mag2(const VECTOR* a)
{
	return vector_dot(a, a);
}

VECTOR*
vector_reflect(VECTOR* o, const VECTOR* N, const VECTOR* E)
{
	float s;
	s = 2 * vector_dot(E, N);
	return vector_init(o,
		s * N->x - E->x,
		s * N->y - E->y,
		s * N->z - E->z);
}

/*****************************************************************************
 *	s
 *****************************************************************************/

/* L@̒` */
#define A00	a->a00
#define A01	a->a01
#define A02	a->a02
#define A03	a->a03
#define A10	a->a10
#define A11	a->a11
#define A12	a->a12
#define A13	a->a13
#define A20	a->a20
#define A21	a->a21
#define A22	a->a22
#define A23	a->a23
#define B00	b->a00
#define B01	b->a01
#define B02	b->a02
#define B03	b->a03
#define B10	b->a10
#define B11	b->a11
#define B12	b->a12
#define B13	b->a13
#define B20	b->a20
#define B21	b->a21
#define B22	b->a22
#define B23	b->a23

const MATRIX MATRIX_I = {
	1, 0, 0, 0,
	0, 1, 0, 0,
	0, 0, 1, 0,
};

MATRIX*
matrix_init(MATRIX* o,
	float a00, float a01, float a02, float a03,
	float a10, float a11, float a12, float a13,
	float a20, float a21, float a22, float a23)
{
	o->a00 = a00; o->a01 = a01; o->a02 = a02; o->a03 = a03;
	o->a10 = a10; o->a11 = a11; o->a12 = a12; o->a13 = a13;
	o->a20 = a20; o->a21 = a21; o->a22 = a22; o->a23 = a23;
	return o;
}

MATRIX*
matrix_xpose(MATRIX* o, const MATRIX* a)
{
	return matrix_init(o,
		A00, A10, A20, 0,
		A01, A11, A21, 0,
		A02, A12, A22, 0);
}

MATRIX*
matrix_inv(MATRIX* o, const MATRIX* a)
{
	int i, j, k;
	float s, t;

	/* 4x4s񂪉24x8s肵A
	 * sAEPʍsƂ܂B
	 */
	float m[4][8] = {
		{ A00, A01, A02, A03, 1, 0, 0, 0 },
		{ A10, A11, A12, A13, 0, 1, 0, 0 },
		{ A20, A21, A22, A23, 0, 0, 1, 0 },
		{   0,   0,   0,   1, 0, 0, 0, 1 },
	};

	/* sim肳鏈... */
	for(i = 0; i < 4; i++) {
		/* m[j=i`3][i]̐Βlԑ傫sjT܂B
		 * _́A0ȊOȂǂł\Ȃ̂łA
		 * 덷邽߂ɁAԑ傫lg܂B
		 */
		j = i;
		s = fabs(m[j][i]);
		for(k = j + 1; k < 4; k++) {
			t = fabs(m[k][i]);
			if(t > s) {
				j = k;
				s = t;
			}
		}
		/* t=0ȂstłAs܂B */

		/* sij܂B */
		if(i != j) {
			for(k = 0; k < 8; k++) {
				s = m[i][k];
				    m[i][k] = m[j][k];
				              m[j][k] = s;
			}
		}

		/* m[i][i]1ɂȂ悤Asim[i][i]Ŋ܂B */
		s = 1.0F / m[i][i];
		for(j = 0; j < 8; j++) m[i][j] *= s;

		/* m[j=iȊO][i]0ɂȂ悤Asim[j][i]{Ĉ܂B */
		for(j = 0; j < 4; j++) {
			if(j == i) continue;
			s = m[j][i];
			for(k = 0; k < 8; k++) m[j][k] -= m[i][k] * s;
		}
		/* ōsim肵܂B */
	}

	/* Es̋tsɂȂĂ܂B */
	return matrix_init(o,
		m[0][4], m[0][5], m[0][6], m[0][7],
		m[1][4], m[1][5], m[1][6], m[1][7],
		m[2][4], m[2][5], m[2][6], m[2][7]);
}

/* mF */
MATRIX*
matrix_invno(MATRIX* o, const MATRIX* a)
{
	/* Ks̋tΐA̒ʂłB
	 *	| R P |-1 = | RT RT*(-P) |
	 * A
	 *	R  = ]\3x3s
	 *	P  = sړ\3xNg
	 *	RT = R̓]us
	 * Ƃ܂B
	 */
	return matrix_init(o,
		A00, A10, A20, -(A00 * A03 + A10 * A13 + A20 * A23),
		A01, A11, A21, -(A01 * A03 + A11 * A13 + A21 * A23),
		A02, A12, A22, -(A02 * A03 + A12 * A13 + A22 * A23));
}

VECTOR*
vector_xform(VECTOR* o, const MATRIX* a, const VECTOR* b)
{
	return vector_init(o,
		A00 * Bx + A01 * By + A02 * Bz + A03,
		A10 * Bx + A11 * By + A12 * Bz + A13,
		A20 * Bx + A21 * By + A22 * Bz + A23);
}

MATRIX*
matrix_xform(MATRIX* o, const MATRIX* a, const MATRIX* b)
{
	return matrix_init(o,
		A00 * B00 + A01 * B10 + A02 * B20,
		A00 * B01 + A01 * B11 + A02 * B21,
		A00 * B02 + A01 * B12 + A02 * B22,
		A00 * B03 + A01 * B13 + A02 * B23 + A03,
		A10 * B00 + A11 * B10 + A12 * B20,
		A10 * B01 + A11 * B11 + A12 * B21,
		A10 * B02 + A11 * B12 + A12 * B22,
		A10 * B03 + A11 * B13 + A12 * B23 + A13,
		A20 * B00 + A21 * B10 + A22 * B20,
		A20 * B01 + A21 * B11 + A22 * B21,
		A20 * B02 + A21 * B12 + A22 * B22,
		A20 * B03 + A21 * B13 + A22 * B23 + A23);
}

MATRIX*
matrix_xform12(MATRIX* o, const MATRIX* a,
	float b00, float b01, float b02, float b03,
	float b10, float b11, float b12, float b13,
	float b20, float b21, float b22, float b23)
{
	return matrix_init(o,
		A00 * b00 + A01 * b10 + A02 * b20,
		A00 * b01 + A01 * b11 + A02 * b21,
		A00 * b02 + A01 * b12 + A02 * b22,
		A00 * b03 + A01 * b13 + A02 * b23 + A03,
		A10 * b00 + A11 * b10 + A12 * b20,
		A10 * b01 + A11 * b11 + A12 * b21,
		A10 * b02 + A11 * b12 + A12 * b22,
		A10 * b03 + A11 * b13 + A12 * b23 + A13,
		A20 * b00 + A21 * b10 + A22 * b20,
		A20 * b01 + A21 * b11 + A22 * b21,
		A20 * b02 + A21 * b12 + A22 * b22,
		A20 * b03 + A21 * b13 + A22 * b23 + A23);
}

MATRIX*
matrix_xlate(MATRIX* o, const MATRIX* a, float x, float y, float z)
{
	return matrix_xform12(o, a,
		1, 0, 0, x,
		0, 1, 0, y,
		0, 0, 1, z);
}

MATRIX*
matrix_scale(MATRIX* o, const MATRIX* a, float x, float y, float z)
{
	return matrix_xform12(o, a,
		x, 0, 0, 0,
		0, y, 0, 0,
		0, 0, z, 0);
}

MATRIX*
matrix_scale1(MATRIX* o, const MATRIX* a, float s)
{
	return matrix_xform12(o, a,
		s, 0, 0, 0,
		0, s, 0, 0,
		0, 0, s, 0);
}

MATRIX*
matrix_rot(MATRIX* o, const MATRIX* a, float x, float y, float z, float d)
{
	VECTOR v;
	float s, c;
	float C;
	/* ]xNg𐳋K܂B */
	vector_init(&v, x, y, z);
	vector_normal(&v, &v);
	x = v.x, y = v.y, z = v.z;
	s = sin(d), c = cos(d);
	C = 1 - c;
	return matrix_xform12(o, a,
		x * x * C +     c, x * y * C - z * s, x * z * C + y * s, 0,
		y * x * C + z * s, y * y * C +     c, y * z * C - x * s, 0,
		z * x * C - y * s, z * y * C + x * s, z * z * C +     c, 0);
}

MATRIX*
matrix_rotx(MATRIX* o, const MATRIX* a, float d)
{
	/* X]śA̒ʂłB
	 *	|  1, 0, 0, 0 |
	 *	|  0, c,-s, 0 |
	 *	|  0, s, c, 0 |
	 * A
	 *	c = cos(d)
	 *	s = sin(d)
	 *	d = ]px
	 * Ƃ܂B
	 */
	float c, s;
	c = cos(d);
	s = sin(d);
	return matrix_xform12(o, a,
		 1, 0, 0, 0,
		 0, c,-s, 0,
		 0, s, c, 0);
}

MATRIX*
matrix_roty(MATRIX* o, const MATRIX* a, float d)
{
	/* Y]śA̒ʂłB
	 *	|  c, 0, s, 0 |
	 *	|  0, 1, 0, 0 |
	 *	| -s, 0, c, 0 |
	 * A
	 *	c = cos(d)
	 *	s = sin(d)
	 *	d = ]px
	 * Ƃ܂B
	 */
	float c, s;
	c = cos(d);
	s = sin(d);
	return matrix_xform12(o, a,
		 c, 0, s, 0,
		 0, 1, 0, 0,
		-s, 0, c, 0);
}

MATRIX*
matrix_rotz(MATRIX* o, const MATRIX* a, float d)
{
	/* Z]śA̒ʂłB
	 *	|  c,-s, 0, 0 |
	 *	|  s, c, 0, 0 |
	 *	|  0, 0, 1, 0 |
	 * A
	 *	c = cos(d)
	 *	s = sin(d)
	 *	d = ]px
	 * Ƃ܂B
	 */
	float c, s;
	c = cos(d);
	s = sin(d);
	return matrix_xform12(o, a,
		 c,-s, 0, 0,
		 s, c, 0, 0,
		 0, 0, 1, 0);
}

MATRIX*
matrix_rotxyx(MATRIX* o, const MATRIX* a, float x, float y, float z)
{
	float cA = cos(x), sA = sin(x);
	float cB = cos(y), sB = sin(y);
	float cC = cos(z), sC = sin(z);
	return matrix_xform12(o, a,
		  cB   ,          sA*sB     ,           cA*sB     , 0,
		-sB*-sC, (cA*cC + sA*cB*-sC), (-sA*cC + cA*cB*-sC), 0,
		-sB*cC , (cA*sC + sA*cB* cC), (-sA*sC + cA*cB* cC), 0);
}

MATRIX*
matrix_rotxzx(MATRIX* o, const MATRIX* a, float x, float y, float z)
{
	float cA = cos(x), sA = sin(x);
	float cB = cos(y), sB = sin(y);
	float cC = cos(z), sC = sin(z);
	return matrix_xform12(o, a,
		cB   ,  cA*-sB             ,  -sA*-sB             , 0,
		sB*cC, (cA* cB*cC + sA*-sC), (-sA* cB*cC + cA*-sC), 0,
		sB*sC, (cA* cB*sC + sA* cC), (-sA* cB*sC + cA* cC), 0);
}

MATRIX*
matrix_rotyxy(MATRIX* o, const MATRIX* a, float x, float y, float z)
{
	float cA = cos(x), sA = sin(x);
	float cB = cos(y), sB = sin(y);
	float cC = cos(z), sC = sin(z);
	return matrix_xform12(o, a,
		(cA* cC + -sA* cB*sC), sB*sC, (sA* cC + cA* cB*sC), 0,
		          -sA*-sB    , cB   ,           cA*-sB    , 0,
		(cA*-sC + -sA* cB*cC), sB*cC, (sA*-sC + cA* cB*cC), 0);
}

MATRIX*
matrix_rotyzy(MATRIX* o, const MATRIX* a, float x, float y, float z)
{
	float cA = cos(x), sA = sin(x);
	float cB = cos(y), sB = sin(y);
	float cC = cos(z), sC = sin(z);
	return matrix_xform12(o, a,
		(cA*cB* cC + -sA*sC), -sB *cC, (sA*cB* cC + cA*sC), 0,
		 cA*sB              ,  cB    ,  sA*sB             , 0,
		(cA*cB*-sC + -sA*cC), -sB*-sC, (sA*cB*-sC + cA*cC), 0);
}

MATRIX*
matrix_rotzxz(MATRIX* o, const MATRIX* a, float x, float y, float z)
{
	float cA = cos(x), sA = sin(x);
	float cB = cos(y), sB = sin(y);
	float cC = cos(z), sC = sin(z);
	return matrix_xform12(o, a,
		(cA*cC + sA*cB*-sC), (-sA*cC + cA*cB*-sC), -sB*-sC, 0,
		(cA*sC + sA*cB* cC), (-sA*sC + cA*cB* cC), -sB* cC, 0,
		         sA*sB     ,           cA*sB     ,  cB    , 0);
}

MATRIX*
matrix_rotzyz(MATRIX* o, const MATRIX* a, float x, float y, float z)
{
	float cA = cos(x), sA = sin(x);
	float cB = cos(y), sB = sin(y);
	float cC = cos(z), sC = sin(z);
	return matrix_xform12(o, a,
		(cA* cB*cC + sA*-sC), (-sA* cB*cC + cA*-sC), sB*cC, 0,
		(cA* cB*sC + sA* cC), (-sA* cB*sC + cA* cC), sB*sC, 0,
		 cA*-sB             ,  -sA*-sB             , cB   , 0);
}

MATRIX*
matrix_rotxyz(MATRIX* o, const MATRIX* a, float x, float y, float z)
{
	float cA = cos(x), sA = sin(x);
	float cB = cos(y), sB = sin(y);
	float cC = cos(z), sC = sin(z);
	return matrix_xform12(o, a,
		 cB*cC, (sA*sB*cC + cA*-sC), (cA*sB*cC + -sA*-sC), 0,
		 cB*sC, (sA*sB*sC + cA* cC), (cA*sB*sC + -sA* cC), 0,
		-sB   ,  sA*cB             ,  cA*cB              , 0);
}

MATRIX*
matrix_rotxzy(MATRIX* o, const MATRIX* a, float x, float y, float z)
{
	float cA = cos(x), sA = sin(x);
	float cB = cos(y), sB = sin(y);
	float cC = cos(z), sC = sin(z);
	return matrix_xform12(o, a,
		cB* cC, (cA*-sB* cC + sA*sC), (-sA*-sB* cC + cA*sC), 0,
		sB    ,  cA* cB             ,  -sA* cB             , 0,
		cB*-sC, (cA*-sB*-sC + sA*cC), (-sA*-sB*-sC + cA*cC), 0);
}

MATRIX*
matrix_rotyxz(MATRIX* o, const MATRIX* a, float x, float y, float z)
{
	float cA = cos(x), sA = sin(x);
	float cB = cos(y), sB = sin(y);
	float cC = cos(z), sC = sin(z);
	return matrix_xform12(o, a,
		(cA*cC + -sA*-sB*-sC), cB*-sC, (sA*cC + cA*-sB*-sC), 0,
		(cA*sC + -sA*-sB* cC), cB* cC, (sA*sC + cA*-sB* cC), 0,
		         -sA* cB     , sB    ,          cA* cB     , 0);
}

MATRIX*
matrix_rotyzx(MATRIX* o, const MATRIX* a, float x, float y, float z)
{
	float cA = cos(x), sA = sin(x);
	float cB = cos(y), sB = sin(y);
	float cC = cos(z), sC = sin(z);
	return matrix_xform12(o, a,
		 cA*cB              , -sB   ,  sA*cB             , 0,
		(cA*sB*cC + -sA*-sC),  cB*cC, (sA*sB*cC + cA*-sC), 0,
		(cA*sB*sC + -sA* cC),  cB*sC, (sA*sB*sC + cA* cC), 0);
}

MATRIX*
matrix_rotzxy(MATRIX* o, const MATRIX* a, float x, float y, float z)
{
	float cA = cos(x), sA = sin(x);
	float cB = cos(y), sB = sin(y);
	float cC = cos(z), sC = sin(z);
	return matrix_xform12(o, a,
		(cA* cC + sA*sB*sC), (-sA* cC + cA*sB*sC),  cB*sC, 0,
		          sA*cB    ,            cA*cB    , -sB   , 0,
		(cA*-sC + sA*sB*cC), (-sA*-sC + cA*sB*cC),  cB*cC, 0);
}

MATRIX*
matrix_rotzyx(MATRIX* o, const MATRIX* a, float x, float y, float z)
{
	float cA = cos(x), sA = sin(x);
	float cB = cos(y), sB = sin(y);
	float cC = cos(z), sC = sin(z);
	return matrix_xform12(o, a,
		          cA* cB     ,          -sA* cB     , sB    , 0,
		(sA*cC +  cA*-sB*-sC), (cA*cC + -sA*-sB*-sC), cB*-sC, 0,
		(sA*sC +  cA*-sB* cC), (cA*sC + -sA*-sB* cC), cB* cC, 0);
}

/* mF */
float
matrix_rotget(VECTOR* o, const MATRIX* a)
{
	float s, t;
	s = acos((A00 + A11 + A22 - 1) / 2);
	if(s != 0) {
		t = 2 * sin(s);
		vector_init(o,
			(A21 - A12) / t,
			(A02 - A20) / t,
			(A10 - A01) / t);
	} else {
		*o = VECTOR_Z;
	}
	return s;
}

MATRIX*
matrix_orient(MATRIX* o, const MATRIX* a,
	float dx, float dy, float dz,
	float ux, float uy, float uz)
{
	/* pśA̒ʂłB
	 *	| rx, ux, dx, 0 |
	 *	| ry, uy, dy, 0 |
	 *	| rz, uz, dz, 0 |
	 * A
	 *	| rx, ry, rz |T = XpxNg
	 *	| ux, uy, uz |T = YpxNg
	 *	| dx, dy, dz |T = ZpxNg
	 * Ƃ܂B
	 */
	VECTOR r, u, d;

	/* ZxNg𐳋K܂B */
	vector_init(&d, dx, dy, dz);
	vector_normal(&d, &d);

	/* XxNg߂āAK܂B 
	 * XxNǵAY->ZƂisA
	 * Ȃ킿OςłB
	 */
	vector_init(&u, ux, uy, uz);
	vector_cross(&r, &u, &d);
	vector_normal(&r, &r);

	/* YxNg߂܂B
	 * YxNǵAZ->XƂisA
	 * Ȃ킿OςłB
	 * ZXxNgɐKĂ̂ŁA
	 * YxNg𖾎IɐKKv͂܂B
	 */
	vector_cross(&u, &d, &r);

	return matrix_xform12(o, a,
		r.x, u.x, d.x, 0,
		r.y, u.y, d.y, 0,
		r.z, u.z, d.z, 0);
}

MATRIX*
matrix_lookat(MATRIX* o, const MATRIX* a,
	float ex, float ey, float ez,
	float cx, float cy, float cz,
	float bank)
{
	/* Jg̓\ϊśA̒ʂłB
	 * 1. oNB
	 * 2. ̊pxA_W璍_Wւ̌ƂB
	 * 3. _WֈړB
	 * _ϊśAJŒ肵ĐE𓮂̂ŁA
	 * 1`3̋tϊƂȂ܂B
	 * s]u^ɂ̂ŁAIuWFNgɋ߂ϊɂȂ܂B
	 *   āA1`3̏͋tōs܂B(matrix_lookat()Ɍ炸)
	 */
	MATRIX m;

	/* _ϊs쐬܂B
	 * ! o==ả\̂ŁAo[NƂĎgĂ͂܂B
	 */
	matrix_xlate (&m, &MATRIX_I, ex, ey, ez);
	matrix_orient(&m, &m, cx - ex, cy - ey, cz - ez, 0, 1, 0);
	matrix_rotz  (&m, &m, bank);
	matrix_invno (&m, &m);

	/* ϊsɊ|킹܂B */
	return matrix_xform(o, a, &m);
}

MATRIX*
matrix_frustum(MATRIX* o, const MATRIX* a, float width, float height, float front)
{
	/* eśA̒ʂłB
	 *	| 2*front/width,              0,                 0,                        0 |
	 *	|             0, 2*front/height,                 0,                        0 |
	 *	|             0,              0, back/(back-front), -front*back/(back-front) |
	 *	|             0,              0,                 1,                        0 |
	 * AłZW𐳋KȂ̂ŁA
	 *	| 2*front/width,              0, 0, 0 |
	 *	|             0, 2*front/height, 0, 0 |
	 *	|             0,              0, 1, 0 |
	 * Ƃ܂BA
	 *	width  = eʂ̕
	 *	height = eʂ̍
	 *	front  = tgNbsO
	 *	back   = obNtbsO
	 * Ƃ܂B
	 */
	return matrix_xform12(o, a,
		2*front/width,              0, 0, 0,
		            0, 2*front/height, 0, 0,
		            0,              0, 1, 0);
}

MATRIX*
matrix_perspect(MATRIX* o, const MATRIX* a, float fovy, float aspect, float front)
{
	float width, height;
	height = front * tan(fovy / 2) * 2; /* eʂ̕   */
	width = height * aspect;            /* eʂ̍ */
	return matrix_frustum(o, a, width, height, front);
}

MATRIX*
matrix_viewport(MATRIX* o, const MATRIX* a, int left, int top, int width, int height)
{
	/* KW(x,y)
	 *	x = :-1`1:E
	 *	y = :-1`1:
	 * Ar[|[gW(X,Y)
	 *	X = :left      `left+width:E
	 *	Y = :top+height`       top:
	 * ւ̕ϊɂāA܂B
	 *
	 * 1) fϊ->_ϊ->ϊ->x=x/z,y=y/z->r[|[gϊ
	 *    Ƃ菇̏ꍇ...
	 * ΐA̒ʂłB
	 *	X = left + (x + 1) *  width  / 2 = ( width  / 2) * x + (left + width  / 2)
	 *	Y = top  + (y - 1) * -height / 2 = (-height / 2) * y + (top  + height / 2)
	 * āAr[|[gϊśA̒ʂłB
	 *	| width/2,         0, 0, left+width/2 |
	 *	|       0, -height/2, 0, top+height/2 |
	 *	|       0,         0, 1,            0 |
	 *
	 * 2) fϊ->_ϊ->ϊ->r[|[gϊ->x=x/z,y=y-z
	 *    Ƃ菇̏ꍇ...
	 * 萔ɂ1/zeĂ܂Ƃ炩ߍlāA
	 *	X = ( width  / 2) * x + (left + width  / 2) * z
	 *	Y = (-height / 2) * y + (top  + height / 2) * z
	 * āAr[|[gϊśA̒ʂłB
	 *	| width/2,         0, left+width/2, 0 |
	 *	|       0, -height/2, top+height/2, 0 |
	 *	|       0,         0,            1, 0 |
	 */
	return matrix_xform12(o, a,
		width/2,         0, 0, left+width/2,
		      0, -height/2, 0, top+height/2,
		      0,         0, 1,            0);
}

MATRIX*
matrix_viewportz(MATRIX* o, const MATRIX* a, int left, int top, int width, int height)
{
	/* matrix_viewport()̃RgQƂĂB */
	return matrix_xform12(o, a,
		width/2,         0, left+width/2, 0,
		      0, -height/2, top+height/2, 0,
		      0,         0,            1, 0);
}

MATRIX*
matrix_shadow(MATRIX* o, const MATRIX* a,
	float Lx, float Ly, float Lz,
	float Gx, float Gy, float Gz,
	float Nx, float Ny, float Nz)
{
	/*	P = eW
	 *	L = xNg
	 *	G = nʂ̈_
	 *	N = nʂ̖@
	 * ɑ΂āAeꂽWS߂@́A̒ʂłB
	 * 炩߁A
	 *	LN = LEN
	 *	GN = GEN
	 *	A  = Lx / LN
	 *	B  = Ly / LN
	 *	C  = Lz / LN
	 * ߂ĂA
	 *	Sx = Px*(1 - Nx*A) + Py*(  - Ny*A) + Pz(  - Nz*A) + GN*A
	 *	Sy = Px*(  - Nx*B) + Py*(1 - Ny*B) + Pz(  - Nz*B) + GN*B
	 *	Sz = Px*(  - Nx*C) + Py*(  - Ny*C) + Pz(1 - Nz*C) + GN*C
	 * āAes
	 *	| 1-Nx*A,  -Ny*A,  -Nz*A, GN*A |
	 *	|  -Nx*B, 1-Ny*B,  -Nz*B, GN*B |
	 *	|  -Nx*C,  -Ny*C, 1-Nz*C, GN*C |
	 * ƂȂ܂B
	 */
	VECTOR L, G, N;
	float LN, GN;
	float A, B, C;
	vector_init(&L, Lx, Ly, Lz);
	vector_init(&G, Gx, Gy, Gz);
	vector_init(&N, Nx, Ny, Nz);
	LN = vector_dot(&L, &N);
	GN = vector_dot(&G, &N);
	A = L.x / LN;
	B = L.y / LN;
	C = L.z / LN;
	return matrix_xform12(o, a,
		1-Nx*A,  -Ny*A,  -Nz*A, GN*A,
		 -Nx*B, 1-Ny*B,  -Nz*B, GN*B,
		 -Nx*C,  -Ny*C, 1-Nz*C, GN*C);
}

float
bezier1(float t, float Q0, float Q1, float Q2)
{
	/* xWFȐ̒`́A̒ʂłB
	 *	P(t) = |t^2 t 1|| 1 -2  1||Q0|
	 *			|-2  2  0||Q1|
	 *			| 1  0  0||Q2|
	 * A
	 *	t: pgbNϐ(0`1)
	 *	Q0: Rg[|Cg#0̍W
	 *	Ql:                     #1
	 *	Q2:                     #2
	 * Ƃ܂B
	 */
	float t2 = t * t;
	return (t2 *  1 + t * -2 + 1 * 1) * Q0 +
	       (t2 * -2 + t *  2 + 1 * 0) * Q1 +
	       (t2 *  1 + t *  0 + 1 * 0) * Q2;
}

float
hermite1(float t, float Q0, float Q1, float Q0d, float Q1d)
{
	/* G~[gȐ̒`́A̒ʂłB
	 *	P(t) = |t^3 t^2 t 1|| 2 -2  1  1||Q0 |
	 *			    |-3  3 -2 -1||Q1 |
	 *			    | 0  0  1  0||Q0'|
	 *			    | 1  0  0  0||Q1'|
	 * A
	 *	t: pgbNϐ(0`1)
	 *	Q0: Rg[|Cg#0̍W
	 *	Ql:                     #1
	 *	Q0': Rg[|Cg#0ɂړ_xNg
	 *	Ql':                     #1
	 * Ƃ܂B
	 */
	float t2 = t  * t;
	float t3 = t2 * t;
	return (t3 *  2 + t2 * -3 + t * 0 + 1 * 1) * Q0  +
	       (t3 * -2 + t2 *  3 + t * 0 + 1 * 0) * Q1  +
	       (t3 *  1 + t2 * -2 + t * 1 + 1 * 0) * Q0d +
	       (t3 *  1 + t2 * -1 + t * 0 + 1 * 0) * Q1d;
}

int
point_in_polygon(float px, float py, float points[/*n_points*/][2], int n_points)
{
	/* * }`ʌ^Ȃ΁Aςg@܂A
	 *   ́A}`ʌ^̏ꍇɂΉł@Ŏ܂B
	 * * ̊֐floatpƂȂĂ܂A̐ɂ悤ɁA
	 *   I[o[t[΍ƂĂ̂ŁA
	 *   ̂܂܉autoϐfloat->intɕύX΁A
	 *   intp̊֐邱Ƃł܂B
	 *   ݂̂ƂA܂intp֐KvȂ߁AĂ܂B
	 *
	 *--------------------------------------------------------------------
	 *
	 * _(px,py)A}`̓ɂ邩ǂ𒲂ׂ@:
	 *
	 * (px,py)n_Ƃ锼ƁA}`̊eӂ̌𒲂ׁA
	 * ӂ̐Ȃ΁A(px,py)͐}`̓łB
	 *
	 * ́A̎ŕ\܂B
	 *	x = px + At	| (A, B): ̌X\xNg
	 *	y = py + Bt	| 0 <= t
	 * }`̂ӂ́A̎ŕ\܂B
	 *	x = x0 + Cu	| C = x1 - x0, D = y1 - y0
	 *	y = y0 + Du	| 0 <= u < 1
	 *			|  0<=u<=1ƂȂŔA
	 *			|    傤ǒ_ʂꍇA
	 *			|    0<=u<=1ł́Aׂ荇̕ӂ̗
	 *			|    Ɣ肳Ă܂łB
	 *			|    ̏ꍇA}`̓炠钸_
	 *			|    ʂĊO֔锼ꍇɁA
	 *			|    }`̓łɂ炸A
	 *			|    _2ƂȂAOƔf܂B
	 *			|    ̊ԈႢh߂ɁA
	 *			|    0<=u<=1ł͂Ȃ0<=u<1ƂĂ܂B
	 * _(x,y)ƂƁA
	 *	x = px + At = x0 + Cu
	 *	y = py + Bt = y0 + Du
	 * AāAt,u߂܂B
	 *	At - Cu = x0 - px
	 *	Bt - Du = y0 - py
	 *	t = (D(x0 - px) - C(y0 - py)) / (AD - BC)
	 *	u = (B(x0 - px) - A(y0 - py)) / (AD - BC)
	 *	t = (DE - CF) / (AD - BC)	| E = x0 - px
	 *	u = (BE - AF) / (AD - BC)	| F = y0 - py
	 * āǍX͔CӂɑIԂƂł̂ŁA
	 * ȒP̂߂ɐƂ܂B
	 *	(A, B) = (1, 0)
	 *	t = (DE - CF) / D
	 *	u = ( 0 -  F) / D
	 * t,ȕA
	 *	0 <= (DE - CF) / D
	 *	0 <= ( 0 -  F) / D < 1
	 * D0ɋ߂ƁAZʂɑ傫ȂAI[o[t[
	 * 댯邽߁AӂDďZȂ܂B
	 *	D >= 0̏ꍇ	| D < 0̏ꍇ
	 *	----------------+-------------
	 *	0 <= DE - CF	| 0 >= DE - CF	... 1
	 *	0 <= -F < D	| 0 >= -F > D	... 2
	 * 1,2ɐ΁Aƕӂ͌܂B
	 * ̔}`̕ӂ̐sAӂȂ
	 * w肳ꂽ_͐}`̓ƂȂ܂B
	 *
	 *--------------------------------------------------------------------
	 */

	int n_cross, i;
	float x0, y0, x1, y1, C, D, E, F;

	if(n_points < 3) return 0;

	/* ܂NAB */
	n_cross = 0;

	/* ŏ̕ӂ̎n_́AŌ̒_Ƃ܂B */
	x0 = points[n_points - 1][0];
	y0 = points[n_points - 1][1];
	for(i = 0; i < n_points; i++) {
		/* ̕ӂ̏I_擾܂B */
		x1 = points[i][0];
		y1 = points[i][1];

		/* B */
		C = x1 - x0;
		D = y1 - y0;
	 	E = x0 - px;
		F = y0 - py;
		if(D >= 0) {
			if(0 <= D * E - C * F &&
			   0 <= -F && -F < D) n_cross++;
		} else {
			if(0 >= D * E - C * F &&
			   0 >= -F && -F > D) n_cross++;
		}

		/* ̕ӂ̎n_́A̕ӂ̏I_łB */
		x0 = x1;
		y0 = y1;
	}

	/* ȂłB */
	return n_cross & 1;
}

float
point_to_line_s(const VECTOR* p, const VECTOR* a, const VECTOR* da)
{
	/* _ p 璼 a + da ֐낵ʒu q ƂB
	 * p  q ւ̃xNgƁAda ͒̂ŁA
	 *	(q - p)Eda = 0		1
	 * q  ɂ̂ŁA
	 *	q = a + da		2
	 * 1,2
	 *	(a + da - p)Eda = 0
	 *	 = daE(p - a) / daEda
	 * : ꕪqɌ da ́APȂZł͂ȂςȂ̂ŁA
	 *       ꕪq da Ŋ (p - a) / da ɂĂ͂ȂB
	 *	 ł́AxNgxNgɂȂĂ܂B
	 *	 ꕪqɋʂ̓ς𑊎E邱Ƃ̂ǂ킩ȂA
	 *	 ȂƂl̓xNgxNǧvZ@mȂB
	 */
	VECTOR ap;
	vector_sub(&ap, p, a);
	return vector_dot(da, &ap) / vector_mag2(da);
}

int
polygon_orientation(float points[/*n_points*/][2], int n_points)
{
	/* }`̋Lq𐳕Ɖ肵āApx̍vꍇA
	 * (}`̊px̍v)ȂA(n_points-2) ƂȂ܂B
	 * t(}`O̊px̍v)ȂA(n_points+2) ƂȂ܂B
	 * ɂ́Av (n_points-2) 𒴂_ŋtƔfł͂łA
	 * Z덷̂߁Ałv (n_point-2)+ ƂȂ\܂B
	 * ŁAŜ߂ɁA҂̒ԂɂlāA
	 * n_points 𒴂_ŋtƔf邱Ƃɂ܂B 
	 *
	 * <<dv!!>
	 * v01v12̊px͂̂ł͂ȂAv10v12̊px͂邱Ƃɒ!
	 * ̕ӂ̕p̍͂̂ł͂ȂA_1̂Ƃ̊px͂̂łB
	 * v01v12̊px͂Ă܂ƁAႦ΋ɒ[ȗȂA
	 * ׂ̗荇ӂ꒼ɂꍇÅԂ̊px͖炩180xłA
	 * Ԉ0xƂȂĂ܂܂B
	 */
	float x0, y0, x1, y1, x2, y2, x10, x12, y10, y12, dx, dy;
	float threshold, d, d_sum;
	int i;

	if(n_points < 3) return 0;

	/* l߂܂B */
	threshold = n_points * PI;

	d_sum = 0;
	for(i = 0; i < n_points; i++) {
		/* v10=(x10,y10), v12=(x12,y12) ߂܂B */
		x0 = points[(i + 0) % n_points][0];
		y0 = points[(i + 0) % n_points][1];
		x1 = points[(i + 1) % n_points][0];
		y1 = points[(i + 1) % n_points][1];
		x2 = points[(i + 2) % n_points][0];
		y2 = points[(i + 2) % n_points][1];
		x10 = x0 - x1;
		y10 = y0 - y1;
		x12 = x2 - x1;
		y12 = y2 - y1;

		/* : dx = |v10||v12|cos */
		dx = x10 * x12 + y10 * y12;

		/* O: dy = |v10||v12|sin */
		dy = x10 * y12 - x12 * y10;

		/* v10v12̊px 0`2 ͈̔͂ŋ߂܂B
		 * : atan2() -΁` ͈̔͂ŕԂė܂B
		 */
		d = atan2(dy, dx);
		if(d < 0) d += PI2;

		/* px𑫂āAl𒴂tłB */
		d_sum += d;
		if(d_sum > threshold) return 1;
	}

	/* Ō܂łl𒴂Ȃ΁AłB */
	return 0;
}

